table.panel-df {
  margin-left: auto;
  margin-right: auto;
  border: none;
  border-collapse: collapse;
  border-spacing: 0;
  font-size: 12px;
  table-layout: auto;
  width: 100%;
}

.panel-df tr,
.panel-df th,
.panel-df td {
  text-align: right;
  vertical-align: middle;
  padding: 0.5em 0.5em;
  line-height: normal;
  white-space: normal;
  max-width: none;
  border: none;
}

.panel-df tbody {
  display: table-row-group;
  vertical-align: middle;
  border-color: inherit;
}

.panel-df tbody tr:nth-child(odd) {
  background: var(--design-surface-color, var(--panel-surface-color));
  color: var(--design-surface-text-color, var(--panel-on-surface-color));
}

.panel-df thead {
  border-bottom: 1px solid
    var(--design-background-text-color, var(--panel-on-background-color));
  vertical-align: bottom;
}

.panel-df tbody tr:hover {
  background-color: #e1f5fe;
  color: #000000;
  cursor: pointer;
}

:host {
  overflow: auto;
  padding-right: 1px;
}
